package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.service.dto.EditorialPageDTO;
import com.vladsch.flexmark.ast.Link;
import com.vladsch.flexmark.ast.ListBlock;
import com.vladsch.flexmark.ast.ListItem;
import com.vladsch.flexmark.ast.Paragraph;
import com.vladsch.flexmark.parser.Parser;
import com.vladsch.flexmark.util.ast.Node;
import com.vladsch.flexmark.util.sequence.BasedSequence;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.WikiPage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cache.annotation.CacheConfig;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;

/**
 * this service enables the provisioning of editorial pages.
 * The basic idea is, to use the wiki of a dedicated git project to edit pages.
 * From there the content is served.
 * @author Michael Breu
 *
 */
@Service
@CacheConfig(cacheNames = { EditorialPagesService.PAGES_CACHE, EditorialPagesService.HELPMENUTREE_CACHE })
public class EditorialPagesService {

    public static final String DE_LANGTAG = "de";

    public static final String EN_LANGTAG = "en";

    private static final Logger LOGGER = LoggerFactory.getLogger(EditorialPagesService.class);

    public static final String PAGES_CACHE = "pages";
    public static final String HELPMENUTREE_CACHE = "helpMenuTree";

    private static final String EDITORIAL_PROJECT_NAME = "infrastructure/GitSearchEditorialPages";

    @Autowired
    @SuppressWarnings("PMD.ImmutableField")
    private GitLabRepository gitLabRepository;

    /**
     * a parser for markup files, in order to construct menu entry.
     */
    private final Parser parser = Parser.builder().build();

    public static class HelpContentTree {

        String title;
        String url;
        List<HelpContentTree> children = new ArrayList<>();

        protected HelpContentTree() {
            // Defaultconstructor for JSON
            super();
        }

        protected HelpContentTree(String title, String url) {
            super();
            this.title = title;
            this.url = url;
        }

        public String getTitle() {
            return title;
        }

        public String getUrl() {
            return url;
        }

        public List<HelpContentTree> getChildren() {
            return children;
        }

        public static HelpContentTree getDefaultTree(String lang) {
            return new HelpContentTree("Help Content", lang + "/helpStart");
        }

        @Override
        public String toString() {
            return toString("");
        }

        public String toString(String prefix) {
            return (
                prefix +
                "[" +
                title +
                "=>" +
                url +
                "]" +
                children.stream().map(m -> "\n" + m.toString(prefix + "  ")).collect(Collectors.joining())
            );
        }
    }

    @Cacheable(PAGES_CACHE)
    public EditorialPageDTO getContent(String path) throws GitLabApiException {
        String encodedPath;
        encodedPath = URLEncoder.encode(path, StandardCharsets.UTF_8);
        final WikiPage page = gitLabRepository.getAdminGitLabApi().getWikisApi().getPage(EDITORIAL_PROJECT_NAME, encodedPath);
        return new EditorialPageDTO(path, page.getContent());
    }

    @CacheEvict(value = PAGES_CACHE, allEntries = true)
    public boolean resetCache() {
        return true; // nothing to do
    }

    @Scheduled(fixedRate = 30 * 60 * 1000L) // every 30 Minutes
    public void evictPagesCacheAtIntervals() {
        resetCache();
    }

    public String getAttachmentURL(String path) {
        final String repositoryURL = gitLabRepository.getRepositoryURL();
        String extraSlash = repositoryURL.endsWith(("/")) ? "" : "/";
        return repositoryURL + extraSlash + "infrastructure/gitsearcheditorialpages/-/wikis/" + path;
    }

    /**
     *
     * @return
     * @throws GitLabApiException
     */
    @Cacheable(HELPMENUTREE_CACHE)
    public HelpContentTree getHelpContentTree(String languageTag) {
        String lang = cleanLanguageTag(languageTag);
        EditorialPageDTO startContent;
        try {
            startContent = this.getContent("/" + lang + "/helpStart");
        } catch (GitLabApiException e) {
            LOGGER.warn("Cannot get pagetree for {}", lang);
            if (EN_LANGTAG.equals(lang)) {
                return HelpContentTree.getDefaultTree(EN_LANGTAG);
            } else {
                // fallback to English
                return getHelpContentTree(EN_LANGTAG);
            }
        }

        Node document = parser.parse(startContent.getContent());

        final HelpContentTree topNode = HelpContentTree.getDefaultTree(lang);

        for (Node firstLevelNode : document.getChildren()) {
            if (firstLevelNode instanceof ListBlock) {
                buildTree((ListBlock) firstLevelNode, topNode);
            }
        }
        LOGGER.debug("TreeNode: {}", topNode);
        return topNode;
    }

    private String cleanLanguageTag(String languageTag) {
        String lang = EN_LANGTAG;
        switch (languageTag) {
            case EN_LANGTAG:
                lang = EN_LANGTAG;
                break;
            case DE_LANGTAG:
                lang = DE_LANGTAG;
                break;
            default:
                lang = EN_LANGTAG;
                break;
        }
        return lang;
    }

    @SuppressWarnings("PMD.AvoidCatchingGenericException")
    private void buildTree(ListBlock listBlock, HelpContentTree parent) {
        try {
            for (Node listElem : listBlock.getChildren()) {
                if (listElem instanceof ListItem) {
                    ListItem listItem = (ListItem) listElem;
                    Node textNode = listItem.getFirstChild();
                    if (textNode instanceof Paragraph) {
                        handleLinks(parent, listElem, (Paragraph) textNode);
                    }
                }
            }
        } catch (Exception e) {
            LOGGER.warn("Problems during parsing HelpTree", e);
            return;
        }
    }

    private void handleLinks(HelpContentTree parent, Node listElem, Paragraph testParagraph) {
        BasedSequence text = testParagraph.getChars();
        Node link = testParagraph.getChildOfType(Link.class);
        HelpContentTree childNode;
        if (link != null) {
            String linkText = link.getFirstChild().getChars().toString().trim();
            String linkUrl = ((Link) link).getPageRef().toString().trim();
            childNode = new HelpContentTree(linkText, linkUrl);
        } else {
            childNode = new HelpContentTree(text.toString().trim(), null);
        }
        parent.getChildren().add(childNode);
        for (Node firstLevelNode : listElem.getChildren()) {
            if (firstLevelNode instanceof ListBlock) {
                buildTree((ListBlock) firstLevelNode, childNode);
            }
        }
    }
}
